﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Gobblet
{
    class Board
    {
        public const int X = 4;
        public const int Y = 4;
        // Board:  1111 = 15 if pieces 1-4
        //         0001 = only piece 1
        //         1001 = piece 4 and piece 1
        //       etc.
        private int[,] boardWhite = new int[X+1, Y]; // x=5 is external stack
        private int[,] boardBlack = new int[X+1, Y];

        public Board()
        {
            for (int x = 0; x < X; x++)
                for (int y = 0; y < Y; y++)
                {
                    boardWhite[x, y] = 0;
                    boardBlack[x, y] = 0;
                }
            for (int y = 0; y < 3; y++)
            {
                boardWhite[X, y] = 4;
                boardBlack[X, y] = 4;
            }

            ////boardWhite[0, 0] = 1;
            //boardWhite[0, 1] = 2;
            //boardWhite[0, 2] = 4;
            //boardWhite[0, 3] = 8;
            //boardBlack[1, 0] = 3;
            ////boardBlack[1, 1] = 6;
            //boardBlack[1, 2] = 5;
            //boardWhite[2, 0] = 10;
            //boardBlack[2, 0] = 5;

            //boardWhite[1, 0] = 8;
            //boardWhite[2, 0] = 8;
            //boardWhite[3, 0] = 8;

            //boardBlack[3, 1] = 1;
            //boardBlack[2, 2] = 1;
            //boardBlack[1, 3] = 8;
            //boardBlack[3, 3] = 4;

            //int[,] ch = find3InRow(frmGobblet.BLACK);
            //for (int x = 0; x < X; x++)
            //{
            //    for (int y = 0; y < Y; y++)
            //    {
            //        Console.Write(ch[x, y]);
            //    }
            //    Console.WriteLine();
            //}
        }

        public int getNumInStackWhite(int num)
        {
            // Should do error checking on input num
            return boardWhite[4,num];
        }

        public int getNumInStackBlack(int num)
        {
            // Should do error checking on input num
            return boardBlack[4,num];
        }

        public int getBlackPiece(int x, int y)
        {
            return boardBlack[x, y];
        }

        public int getWhitePiece(int x, int y)
        {
            return boardWhite[x, y];
        }

        // Gets the color of the piece at the specified coordinates
        public int getColor(int x, int y)
        {
            int white = boardWhite[x,y];
            int black = boardBlack[x, y];
            if ((white == 0) && (black == 0))
            {
                // no piece
                return frmGobblet.NONE;
            }
            else if (white > black)
            {
                return frmGobblet.WHITE;
            }
            else
            {
                return frmGobblet.BLACK;
            }
        }

        // Determines the size of the piece from the board binary representation
        public int decodeBiggestSize(int num)
        {
            int size = 0;
            if ((num / 8) >= 1)
                size = 4;
            else if ((num / 4) >= 1)
                size = 3;
            else if ((num / 2) >= 1)
                size = 2;
            else if (num >= 1)
                size = 1;
            return size;
        }

        public Board(Board newBoard)
        {
            for (int x=0; x<=X; x++)
                for (int y = 0; y < Y; y++)
                {
                    boardWhite[x, y] = newBoard.boardWhite[x, y];
                    boardBlack[x, y] = newBoard.boardBlack[x, y];
                }
        }

        // Computes 2D array of 1's that match locations that
        // are gobblable from an external stack
        // for the specified color        
        /* 
         * This version not correct, assumed 3 in a row had to be touching,
         * but can be non-contiguous in a row
         * 
        public int[,] find3InRow(int c)
        {
            int[,] pieces = new int[X, Y];
            int[,] gobble = new int[X, Y];
            for (int x = 0; x < X; x++)
                for (int y = 0; y < Y; y++)
                {
                    pieces[x, y] = getColor(x, y);
                    gobble[x, y] = 0;
                }

            // Calculate rows
            for (int x = 0; x < X; x++)
            {
                if ((pieces[x, 0] == c) && (pieces[x, 1] == c) && (pieces[x, 2] == c))
                {
                    gobble[x, 0] = 1; gobble[x, 1] = 1; gobble[x, 2] = 1;
                }
                if ((pieces[x, 1] == c) && (pieces[x, 2] == c) && (pieces[x, 3] == c))
                {
                    gobble[x, 1] = 1; gobble[x, 2] = 1; gobble[x, 3] = 1;
                }
            }
            // Calculate columns
            for (int y = 0; y < Y; y++)
            {
                if ((pieces[0, y] == c) && (pieces[1, y] == c) && (pieces[2, y] == c))
                {
                    gobble[0, y] = 1; gobble[0, y] = 1; gobble[0, y] = 1;
                }
                if ((pieces[1, y] == c) && (pieces[2, y] == c) && (pieces[3, y] == c))
                {
                    gobble[1, y] = 1; gobble[2, y] = 1; gobble[3, y] = 1;
                }
            }
            // Calculate diagonals
            if ((pieces[0, 0] == c) && (pieces[1, 1] == c) && (pieces[2, 2] == c))
            {
                gobble[0, 0] = 1; gobble[1, 1] = 1; gobble[2, 2] = 1;
            }
            if ((pieces[3, 3] == c) && (pieces[1, 1] == c) && (pieces[2, 2] == c))
            {
                gobble[3, 3] = 1; gobble[1, 1] = 1; gobble[2, 2] = 1;
            }
            if ((pieces[3, 0] == c) && (pieces[2, 1] == c) && (pieces[1, 2] == c))
            {
                gobble[3, 0] = 1; gobble[2, 1] = 1; gobble[1, 2] = 1;
            }
            if ((pieces[0, 3] == c) && (pieces[2, 1] == c) && (pieces[1, 2] == c))
            {
                gobble[0, 3] = 1; gobble[2, 1] = 1; gobble[1, 2] = 1;
            }     
            // Non-center diagonals
            if ((pieces[1, 0] == c) && (pieces[2, 1] == c) && (pieces[3, 2] == c))
            {
                gobble[1, 0] = 1; gobble[2, 1] = 1; gobble[3, 2] = 1;
            }
            if ((pieces[0, 1] == c) && (pieces[1, 2] == c) && (pieces[2, 3] == c))
            {
                gobble[0, 1] = 1; gobble[1, 2] = 1; gobble[2, 3] = 1;
            }
            if ((pieces[2, 0] == c) && (pieces[1, 1] == c) && (pieces[0, 2] == c))
            {
                gobble[2, 0] = 1; gobble[1, 1] = 1; gobble[0, 2] = 1;
            }
            if ((pieces[3, 1] == c) && (pieces[2, 2] == c) && (pieces[1, 3] == c))
            {
                gobble[3, 1] = 1; gobble[2, 2] = 1; gobble[1, 3] = 1;
            }     

            return gobble;
        }
         **/

        // Computes 2D array of 1's that match locations that
        // are gobblable from an external stack
        // for the specified color        
        public int[,] find3InRow(int c)
        {
            int[,] pieces = new int[X, Y];
            int[,] gobble = new int[X, Y];            
            for (int x = 0; x < X; x++)
                for (int y = 0; y < Y; y++)
                {
                    pieces[x, y] = getColor(x, y);
                    gobble[x, y] = 0;
                }
            int sum;

            // Calculate rows
            for (int x = 0; x < X; x++)
            {
                sum = 0;
                for (int y = 0; y < Y; y++)
                    if (pieces[x, y] == c) sum++;
                if (sum == 3)
                    for (int y = 0; y < Y; y++)
                        if (pieces[x, y] == c) gobble[x, y] = 1;
            }
            // Calculate columns
            for (int y = 0; y < Y; y++)
            {
                sum = 0;
                for (int x = 0; x < X; x++)
                    if (pieces[x, y] == c) sum++;
                if (sum == 3)
                    for (int x = 0; x < X; x++)
                        if (pieces[x, y] == c) gobble[x, y] = 1;
            }
            // Calculate diagonals
            sum = 0;
            for (int x = 0; x < X; x++)
            {
                if (pieces[x, x] == c) sum++;
            }
            if (sum == 3)
                for (int x = 0; x < X; x++)
                    if (pieces[x, x] == c) gobble[x, x] = 1;
            sum = 0;
            for (int x = 0; x < X; x++)
            {
                if (pieces[x, 3-x] == c) sum++;
            }
            if (sum == 3)
                for (int x = 0; x < X; x++)
                    if (pieces[x, 3-x] == c) gobble[x, 3-x] = 1;   
            // Non-center diagonals
            if ((pieces[1, 0] == c) && (pieces[2, 1] == c) && (pieces[3, 2] == c))
            {
                gobble[1, 0] = 1; gobble[2, 1] = 1; gobble[3, 2] = 1;
            }
            if ((pieces[0, 1] == c) && (pieces[1, 2] == c) && (pieces[2, 3] == c))
            {
                gobble[0, 1] = 1; gobble[1, 2] = 1; gobble[2, 3] = 1;
            }
            if ((pieces[2, 0] == c) && (pieces[1, 1] == c) && (pieces[0, 2] == c))
            {
                gobble[2, 0] = 1; gobble[1, 1] = 1; gobble[0, 2] = 1;
            }
            if ((pieces[3, 1] == c) && (pieces[2, 2] == c) && (pieces[1, 3] == c))
            {
                gobble[3, 1] = 1; gobble[2, 2] = 1; gobble[1, 3] = 1;
            }     
            return gobble;
        }

        // Returns white, black, or NONE if no winner
        //
        // BUG:  If move reveals 4 in a row and also gets 4 in a row, 
        // could return the wrong winner
        public int winner()
        {
            int[,] pieces = new int[X, Y];
            int sumblack = 0, sumwhite = 0;

            for (int x = 0; x < X; x++)
                for (int y = 0; y < Y; y++)
                {
                    pieces[x, y] = getColor(x, y);
                }
            // Check rows
            for (int x = 0; x < X; x++)
            {
                sumblack = 0;
                sumwhite = 0;
                for (int y = 0; y < Y; y++)
                {
                    if (pieces[x, y] == frmGobblet.WHITE)
                        sumwhite++;
                    else if (pieces[x, y] == frmGobblet.BLACK)
                        sumblack++;
                    if (sumwhite == 4)
                        return frmGobblet.WHITE;
                    if (sumblack == 4)
                        return frmGobblet.BLACK;
                }               
            }
            // Check columns
            for (int y = 0; y < Y; y++)
            {
                sumblack = 0;
                sumwhite = 0;
                for (int x = 0; x < X; x++)
                {
                    if (pieces[x, y] == frmGobblet.WHITE)
                        sumwhite++;
                    else if (pieces[x, y] == frmGobblet.BLACK)
                        sumblack++;
                    if (sumwhite == 4)
                        return frmGobblet.WHITE;
                    if (sumblack == 4)
                        return frmGobblet.BLACK;
                }
            }
            // Check diagonals
            sumblack = 0;
            sumwhite = 0;
            for (int x = 0; x < X; x++)
            {
                if (pieces[x, x] == frmGobblet.WHITE)
                    sumwhite++;
                else if (pieces[x, x] == frmGobblet.BLACK)
                    sumblack++;
                if (sumwhite == 4)
                    return frmGobblet.WHITE;
                if (sumblack == 4)
                    return frmGobblet.BLACK;
            }
            sumblack = 0;
            sumwhite = 0;
            for (int x = 0; x < X; x++)
            {
                if (pieces[x, 3-x] == frmGobblet.WHITE)
                    sumwhite++;
                else if (pieces[x, 3-x] == frmGobblet.BLACK)
                    sumblack++;
                if (sumwhite == 4)
                    return frmGobblet.WHITE;
                if (sumblack == 4)
                    return frmGobblet.BLACK;
            }
            return frmGobblet.NONE;
        }

        // Actually make the specified move to the board
        public void makeMove(int x1, int y1, int x2, int y2, int whoseturn)
        {
            if (whoseturn == frmGobblet.WHITE)
            {
                // Get size
                int size = 0;
                if (x1 == 4)
                {
                    size = boardWhite[x1, y1];
                }
                else
                {
                    size = decodeBiggestSize(boardWhite[x1, y1]);
                }
                // Remove from source
                int encodedVal = (int)Math.Pow(2, size-1);
                if (x1 == 4)
                    boardWhite[x1, y1]--;
                else
                    boardWhite[x1, y1] -= encodedVal;
                // Add to destination
                boardWhite[x2, y2] += encodedVal;
            }
            else
            {
                // Get size
                int size = 0;
                if (x1 == 4)
                {
                    size = boardBlack[x1, y1];
                }
                else
                {
                    size = decodeBiggestSize(boardBlack[x1, y1]);
                }
                // Remove from source
                int encodedVal = (int)Math.Pow(2, size - 1);
                if (x1 == 4)
                    boardBlack[x1, y1]--;
                else
                    boardBlack[x1, y1] -= encodedVal;
                // Add to destination
                boardBlack[x2, y2] += encodedVal;
            }
        }

        // Determines if the specified move is valid
        public bool validMove(int x1, int y1, int x2, int y2, int whoseturn)
        {
            // Outside allowable range
            if ((x1 < 0) || (x1 > 4) || (y1 < 0) || (y1 > 3) ||
                (x2 < 0) || (x2 > 4) || (y2 < 0) || (y2 > 3))
                return false;
            // Move to same spot
            if ((x1 == x2) && (y1 == y2))
                return false;
            // Try to move a piece that is not yours
            if (x1 < 4)
            {
                int color = getColor(x1, y1);
                if (color != whoseturn)
                    return false;
            }
            // Try to move from a stack that is empty
            if (x1 == 4)
            {
                if (whoseturn == frmGobblet.WHITE)
                {
                    if (boardWhite[x1, y1] < 1) return false;
                }
                else
                {
                    if (boardBlack[x1, y1] < 1) return false;
                }
            }

            // Get size of piece being moved
            int size1 = 0;
            if (x1 == 4)
            {
                if (whoseturn == frmGobblet.WHITE)
                    size1 = boardWhite[x1, y1];
                else
                    size1 = boardBlack[x1, y1];
            }
            else
            {
                if (whoseturn == frmGobblet.WHITE)
                    size1 = decodeBiggestSize(boardWhite[x1, y1]);
                else
                    size1 = decodeBiggestSize(boardBlack[x1, y1]);                
            }

            // Handle external stack to board move
            if (x1 == 4)
            {
                int col = getColor(x2, y2);
                if (col == frmGobblet.NONE)
                {
                    // Valid
                    return true;
                }
                else if (col != whoseturn)
                {
                    // Trying to put an external stack to a piece, valid if 3 in a row
                    // and its the opponent
                    int[,] numrow;
                    if (whoseturn == frmGobblet.WHITE)
                        numrow = find3InRow(frmGobblet.BLACK);
                    else
                        numrow = find3InRow(frmGobblet.WHITE);
                    if (numrow[x2, y2] == 1)
                    {
                        // 3 in row of opponent, valid if size is smaller
                        int size2;
                        if (col == frmGobblet.WHITE)
                            size2 = decodeBiggestSize(boardWhite[x2, y2]);
                        else
                            size2 = decodeBiggestSize(boardBlack[x2, y2]);
                        if (size1 > size2)
                            return true;
                    }
                    return false;
                }
                else
                {
                    return false;
                }
            }
            else
            {
                // Moving from board to board
                int col = getColor(x2, y2);
                if (col == frmGobblet.NONE)
                {
                    return true;
                }
                int size2;
                if (col == frmGobblet.WHITE)
                    size2 = decodeBiggestSize(boardWhite[x2, y2]);
                else
                    size2 = decodeBiggestSize(boardBlack[x2, y2]);
                if (size1 > size2)
                    return true;
                else
                    return false;                
            }
        }

        // Generate movelist
        public List<Move> getMoveList(int whosemove)
        {
            List<Move> moveList = new List<Move>();
            int x, y;
            int[,] pieces = new int[X, Y];

            // Get color of pieces on board
            for (x = 0; x < X; x++)
                for (y = 0; y < Y; y++)
                {
                    pieces[x, y] = getColor(x, y);
                }

            // Generate moves on board
            for (x = 0; x < X; x++)
                for (y = 0; y < Y; y++)
                {
                    if (pieces[x, y] == whosemove)
                    {
                        // try all destination spots
                        for (int x2 = 0; x2 < X; x2++)
                            for (int y2 = 0; y2 < Y; y2++)
                            {
                                if (validMove(x, y, x2, y2, whosemove))
                                    moveList.Add(new Move(x, y, x2, y2));
                            }
                    }
                }
            // Generate moves from external stacks
            x = 4;
            for (y = 0; y < 3; y++)
            {
                // try all destination spots
                for (int x2 = 0; x2 < X; x2++)
                    for (int y2 = 0; y2 < Y; y2++)
                    {
                        if (validMove(x, y, x2, y2, whosemove))
                            moveList.Add(new Move(x, y, x2, y2));
                    }
            }
            return moveList;
        }

    }
}
