﻿/****
 * Gobblet Random Move player.
 * Intended as a demo for generating moves, representing the board, and connecting to the 
 * game manager server.
 */
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Net;
using System.Net.Sockets;
using System.Threading;

namespace Gobblet
{
    public partial class frmGobblet : Form
    {
        public const int WHITE = 0;
        public const int BLACK = 1;
        public const int NONE = -1;

        public const int MAXTIME = 25;      // max # of seconds to make a move

        private TcpClient sockServer;           // Used to connect to the server
        private NetworkStream networkStream;            // Used for buffered I/O to the server
        private System.IO.StreamReader streamReader;
        private System.IO.StreamWriter streamWriter;

        private Random rand = new Random(); // To make random moves
        private int whoseturn = WHITE;
        private int compPlayer = BLACK;
        
        // Internal control array for board
        private Label[,] displayBoard = new Label[4, 4];
        private Label[] displayWhiteStacks = new Label[3];
        private Label[] displayBlackStacks = new Label[3];

        // Actual game board
        Board b = new Board();

        // Start time for move
        private DateTime startTime;

        public frmGobblet()
        {
            InitializeComponent();
        }

        private void showBoard()
        {
            if (whoseturn == WHITE)
                lblInfo.Text = "White's move.";
            else
                lblInfo.Text = "Black's move.";

            for (int x = 0; x < Board.X; x++)
                for (int y = 0; y < Board.Y; y++)
                {
                    int white = b.getWhitePiece(x, y);
                    int black = b.getBlackPiece(x, y);
                    if ((white == 0) && (black == 0))
                    {
                        // no piece
                        displayBoard[x, y].BackColor = Color.LightGray;
                        displayBoard[x, y].Text = "";
                    }
                    else if (white > black)
                    {
                        // white piece
                        // figure out biggest size
                        int size = 0;
                        if ((white / 8) >= 1)
                            size = 4;
                        else if ((white / 4) >= 1)
                            size = 3;
                        else if ((white / 2) >= 1)
                            size = 2;
                        else if (white >= 1)
                            size = 1;
                        displayBoard[x, y].BackColor = Color.White;
                        displayBoard[x, y].ForeColor = Color.Black;
                        displayBoard[x, y].Text = size.ToString();
                    }
                    else
                    {
                        // black piece
                        int size = 0;
                        if ((black / 8) >= 1)
                            size = 4;
                        else if ((black / 4) >= 1)
                            size = 3;
                        else if ((black / 2) >= 1)
                            size = 2;
                        else if (black >= 1)
                            size = 1;
                        displayBoard[x, y].BackColor = Color.Black;
                        displayBoard[x, y].ForeColor = Color.White;
                        displayBoard[x, y].Text = size.ToString();
                    }
                }
            // Show stacks
            for (int y = 0; y < 3; y++)
            {
                int white = b.getNumInStackWhite(y);
                int black = b.getNumInStackBlack(y);
                displayWhiteStacks[y].Text = white.ToString();
                displayBlackStacks[y].Text = black.ToString();
            }
        }

        private void frmGobblet_Load(object sender, EventArgs e)
        {
            // Maps controls to a 2D array so we can update the board display by changing array entries
            displayBoard[0, 0] = lbl00;
            displayBoard[0, 1] = lbl01;
            displayBoard[0, 2] = lbl02;
            displayBoard[0, 3] = lbl03;
            displayBoard[1, 0] = lbl10;
            displayBoard[1, 1] = lbl11;
            displayBoard[1, 2] = lbl12;
            displayBoard[1, 3] = lbl13;
            displayBoard[2, 0] = lbl20;
            displayBoard[2, 1] = lbl21;
            displayBoard[2, 2] = lbl22;
            displayBoard[2, 3] = lbl23;
            displayBoard[3, 0] = lbl30;
            displayBoard[3, 1] = lbl31;
            displayBoard[3, 2] = lbl32;
            displayBoard[3, 3] = lbl33;
            displayWhiteStacks[0] = lbl40white;
            displayWhiteStacks[1] = lbl41white;
            displayWhiteStacks[2] = lbl42white;
            displayBlackStacks[0] = lbl40black;
            displayBlackStacks[1] = lbl41black;
            displayBlackStacks[2] = lbl42black;
            showBoard();
        }

        /*
         * This method currently not used, it was used for human input
         * but the input textbox has been disabled for this program.
         */
        private void txtMove_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == 13)
            {
                int x1, y1, x2, y2;
                String s = txtMove.Text;
                if (s.Length == 4)
                {
                    x1 = s[0] - '0';
                    y1 = s[1] - '0';
                    x2 = s[2] - '0';
                    y2 = s[3] - '0';
                    if (x1 == 5)
                        x1 = 4;
                    else if (x1 == 6)
                        x1 = 4;
                    else
                    {
                        x1--;
                    }
                    y1--;
                    x2--;
                    y2--;
                    MakeMove(x1, y1, x2, y2);
                }

                txtMove.Text = "";
                e.Handled = true;
            }
        }

        private void MakeMove(int x1, int y1, int x2, int y2)
        {
            if (b.validMove(x1, y1, x2, y2, whoseturn))
            {
                b.makeMove(x1, y1, x2, y2, whoseturn);
                AddMoveToList(new Move(x1, y1, x2, y2));
                showBoard();

                CheckForWin();

                // Switch turns
                if (whoseturn == WHITE)
                {
                    whoseturn = BLACK;
                }
                else
                {
                    whoseturn = WHITE;
                }
                showBoard();
                if (whoseturn == compPlayer)
                {
                    ComputerMove();
                }
            }
            else
            {
                MessageBox.Show("Invalid move");
            }
        }

        private bool CheckForWin()
        {
            // Check for win
            int result = b.winner();
            if (result == WHITE)
            {
                lblInfo.BackColor = Color.Black;
                lblInfo.ForeColor = Color.White;
                lblInfo.Text = "White wins!";
                txtMove.Enabled = false;
                return true;
            }
            if (result == BLACK)
            {
                lblInfo.BackColor = Color.Black;
                lblInfo.ForeColor = Color.White;
                lblInfo.Text = "Black wins!";
                txtMove.Enabled = false;
                return true;
            }
            return false;
        }

        /*
         * This method just picks a random move from the board and returns it
         * as a string.
         */
        private string ComputerMove()
        {
            lblInfo.Text = "Computer thinking...";
            Cursor.Current = Cursors.WaitCursor; 
            Application.DoEvents();

            //Make random move
            List<Move> moveList = b.getMoveList(whoseturn);
            int i = rand.Next(moveList.Count);
            Move m = moveList[i];

            b.makeMove(m.X1, m.Y1, m.X2, m.Y2, whoseturn);
            AddMoveToList(m);
            showBoard();
            Cursor.Current = Cursors.Default; 

            return m.getMove(whoseturn);
        }

        // Displays move in a listbox, scrolls to end
        private void AddMoveToList(Move m)
        {
            String s = m.getMove(whoseturn);
            lstMoves.Items.Add(s);
            lstMoves.TopIndex = lstMoves.Items.Count - 1;
        }

        private void quitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            MessageBox.Show("This program makes random moves and is just a demo of connecting to the server.","Dumb Gobblet AI");
        }

        /**
         * This method connects to the server on port 9999.
         * It uses the protocol of sending PLAYER then our name and waiting for
         * what player we are and the name of the opponent.
         */
        private void btnConnect_Click(object sender, EventArgs e)
        {
            this.sockServer = new TcpClient(txtIPAddress.Text, 9999);
            networkStream = sockServer.GetStream();
            streamReader = new System.IO.StreamReader(networkStream);
            streamWriter = new System.IO.StreamWriter(networkStream);

            // Send PLAYER
            sendLine("PLAYER");
            // Send our player name
            string name = txtPlayerName.Text;
            sendLine(name);

            string s = recvLine();      // "PLAYER"
            s = recvLine();             // Player num
            if (s.Equals("1"))
            {
                compPlayer = WHITE;
                this.Text = "Gobblenator - Player 1";
            }
            else
            {
                compPlayer = BLACK;
                this.Text = "Gobblenator - Player 2";
            }
            Application.DoEvents();
            s = recvLine();             // Player1/2Name
            s = recvLine();             // Opponent's name     - This program just ignore these
            playGame();                 // Plays the game, waiting for turns
        }

        // Send a move to the server
        private void sendLine(string message)
        {
            try
            {
                streamWriter.WriteLine(message);
                streamWriter.Flush();
            }
            catch (Exception e)
            {
                MessageBox.Show(e.ToString());
            }
            Thread.Sleep(250);      // Needed this to ensure data sent? flush doesn't send?
        }

        // Receive a move from the server.  The server
        // appears to return strings terminated in \r\n.
        private string recvLine()
        {
            string s = streamReader.ReadLine().Trim();
            // Check for WIN/LOSE here, otherwise the rest of the program
            // tries to parse the string as a numeric move
            if (s.Equals("LOSE"))
            {
                MessageBox.Show("You lose, you made an invalid move?");
                this.Close();
            }
            else if (s.Equals("WIN"))
            {
                MessageBox.Show("You win, the opponent made an invalid move.");
                this.Close();
            }
            return s;
        }

        // Master loop to play the game.
        // Runs as white first, then black, and continues until the
        // game is over.
        private void playGame()
        {
            string move;
            int x1, x2, y1, y2;

            while (true)
            {
                // White's turn
                whoseturn = WHITE;
                showBoard();
                Application.DoEvents();
                if (compPlayer == WHITE)
                {
                    move = ComputerMove();      // Makes the move on the board
                    sendLine(move); // Send our move to the server
                }
                else
                {
                    move = recvLine(); // Not our move, wait for server to tell us what the opponent did
                    lstMoves.Items.Add(move);
                    lstMoves.TopIndex = lstMoves.Items.Count - 1;
                    // Parse move and make it on the board
                    x1 = move[0]- '0';
                    y1 = move[1] - '0';
                    x2 = move[2] - '0';
                    y2 = move[3] - '0';      
                    // Map move to internal move format
                    if (x1 == 5)
                        x1 = 4;
                    else if (x1 == 6)
                        x1 = 4;
                    else
                    {
                        x1--;
                    }
                    y1--;
                    x2--;
                    y2--;              
                    b.makeMove(x1,y1,x2,y2, WHITE);
                }
                if (CheckForWin())
                    return;
                Application.DoEvents();

                // Black's turn
                whoseturn = BLACK;
                showBoard();
                Application.DoEvents();
                if (compPlayer == BLACK)
                {
                    move = ComputerMove();      // Makes the move on the board
                    sendLine(move); // Send our move to the server
                }
                else
                {
                    move = recvLine(); // Not our move, wait for server to tell us what the opponent did
                    lstMoves.Items.Add(move);
                    lstMoves.TopIndex = lstMoves.Items.Count - 1;
                    // Parse move and make it on the board
                    x1 = move[0] - '0';
                    y1 = move[1] - '0';
                    x2 = move[2] - '0';
                    y2 = move[3] - '0';
                    // Map move format to the internal move format
                    if (x1 == 5)
                        x1 = 4;
                    else if (x1 == 6)
                        x1 = 4;
                    else
                    {
                        x1--;
                    }
                    y1--;
                    x2--;
                    y2--;
                    b.makeMove(x1, y1, x2, y2, BLACK);
                }
                if (CheckForWin())
                    return;
                Application.DoEvents();
            }
        }


    }

}
